from __future__ import division, print_function, absolute_import
import numpy as np
from copy import copy
from .saft_forcefield import saft_forcefield
from .constants import kb
# used in saftgammamie
from .config_asso import asso_aux
from .secondorder import secondorder47_1, secondorder47_2, secondorder19_14
from .secondorder import secondorder21, secondorder22, secondorder23
from .database import database


class component(object):
    '''
    Creates an object with pure component info

    Parameters
    ----------
    name : str
        Name of the component
    ms : float
        Chain lenght
    sigma : float
        diameter, input in [Ángstrom], stored in [m]
    eps : float
        Mie Potential energy, input in [k], stored in [J]
    lambda_a : float
        Atractive exponent of Mie Potential
    lambda_r : float
        Repulsive exponent of Mie Potential
    eAB : float
        Association Energy, input in [k], stores in [J]
    rcAB : float
        Association rc, input in [sigma], stored in [m]
    rdAB : float
        Association rd, input in [sigma], stored in [m]
    sites : list
        Association sites [Bipolar, Positive, Negative]
    mupol : float
        dipolar moment [Debye]
    npol : float
        numer of polar sites de GV polar term
    ring : float
        geometric factor for ring molecules
        (see Langmuir 2017, 33, 11518-11529, Table I.)
    Tc : float
        Critical temperature [K]
    Pc : float
        Critical Pressure [Pa]
    Zc : float
        critical compresibility factor
    Vc : float
        critical volume [m^3/mol]
    w  : float
        acentric factor
    Mw : float
        molar weight [g/mol]
    cii : list
        polynomial coefficient for influence parameter in SGT [J m^5 / mol^2]

    Methods
    -------
    ci :  evaluates influence parameter polynomial
    saftvrmie_forcefield: use corresponding state principle to correlate
        sigma, eps and lambda_a.
    '''

    def __init__(self, name='None', ms=1., sigma=0., eps=0., lambda_r=12.,
                 lambda_a=6., eAB=0., rcAB=1., rdAB=0.4, sites=[0, 0, 0],
                 mupol=0, npol=0., ring=0., Tc=0., Pc=0., Zc=0., Vc=0.,
                 w=0., Mw=1., cii=0., GC=None):

        self.name = name
        self.Tc = Tc  # Critical Temperature in K
        self.Pc = Pc  # Critical Pressure in Pa
        self.Zc = Zc  # Critical compresibility factor
        self.Vc = Vc  # Critical volume in m3/mol
        self.w = w  # Acentric Factor
        self.cii = np.atleast_1d(cii)  # Influence factor SGT, list or array
        self.Mw = Mw  # molar weight in g/mol
        self.nc = 1
        self.GC = GC  # Dict, Group contribution info

        # Saft Parameters
        self.ms = ms
        self.sigma = sigma * 1e-10  # meters
        self.eps = eps * kb  # Joule
        self.lambda_a = np.asarray(lambda_a, dtype=float)
        self.lambda_r = np.asarray(lambda_r, dtype=float)
        self.lambda_ar = self.lambda_r + self.lambda_a

        # For ring molecules (see Langmuir 2017, 33, 11518-11529, Table I.)
        self.ring = ring

        # Association Parameters
        self.eAB = eAB * kb  # Joule
        self.rcAB = rcAB * self.sigma  # meters
        self.rdAB = rdAB * self.sigma  # meters
        self.sites = sites

        # Polar parameters
        self.mupol = mupol  # Debye
        self.npol = npol

    def __add__(self, component2):
        '''
        Methods to add two components and create a mixture with them.
        '''
        return mixture(self, component2)

    def ci(self, T):
        """
        Method that evaluates the polynomial for cii coeffient of SGT
        cii must be in J m^5 / mol^2 and T in K.

        Parameters
        ----------
        T : float
            absolute temperature [K]

        Returns
        -------
        ci : float
            influence parameter at given temperature [J m^5 / mol^2]

        """

        return np.polyval(self.cii, T)

    def saftvrmie_forcefield(self, ms, rhol07, ring_type=None):
        """
        saftvrmie_forcefield(ms, rhol07)

        Method that uses corresponding state principle to correlate sigma,
        eps and lambda_r using critical temperature, acentric factor and
        liquid density (Tr=0.7)

        This method requires that critical temperature and acentric factor of
        the fluid has been set.

        This method overwrites the following attributes: ms, sigma, eps,
        lambda_r, lambda_a.

        Parameters
        ----------
        ms : integrer
            number of chain segments
        rhol07: float
            liquid density at reduced temperature = 0.7 [mol/m^3]
        ring_type : None or string
            ring type, 'None' for chain molecules, ring: 'a' for ms=3,
            'b' for ms=4, 'c' or 'd' for ms=5 and 'e' or 'f' for ms=7.
            See Table 1 of Langmuir 2017 33 (42), 11518-11529 or
            Table S2, Supporting information 2 of J. Chem. Inf. Model.
            2021, 61, 3, 1244–1250 for more information about ring geometries.

        Returns
        -------
        sigma : float,
            size parameter for Mie potential [m]
        eps : float
            well depth for Mie potential [K]
        lambda_r : float
            repulsive exponent for Mie potential [Adim]
        ring : float
            geometric factor for ring molecules

        """
        out = saft_forcefield(ms, self.Tc, self.w, rhol07, ring_type)
        lambda_r, lambda_a, ms, eps, sigma, ring = out
        self.lambda_a = np.asarray(lambda_a)
        self.lambda_r = np.asarray(lambda_r)
        self.lambda_ar = self.lambda_r + self.lambda_a
        self.ms = ms
        self.sigma = sigma
        self.eps = eps
        self.ring = ring
        return sigma, eps/kb, lambda_r, ring

    def saftgammamie(self, database=database):
        """
        saftgammamie method

        Method that reads the SAFT-Gamma-Mie database and set the necessary
        interactions parameters to use the equation of state.

        Parameters
        ----------
        database: Object
            database object
        """
        df_groups = database.df_groups
        df_mie_kl = database.df_mie_kl
        df_asso_kl = database.df_asso_kl
        df_secondorder = database.df_secondorder

        GC = self.GC
        nc = self.nc

        subgroups = np.asarray(list(GC.keys()))
        ngroups = len(subgroups)
        vki = np.asarray(list(GC.values()))
        ngtotal = np.sum(ngroups)
        groups_index = np.zeros(ngtotal, dtype=np.int64)
        group_indexes = [[0, ngtotal]]

        group_filter = df_groups.loc[subgroups]

        vk = np.asarray(group_filter['vk*'])
        Sk = np.asarray(group_filter['Sk'])

        sigma_kk = np.asarray(group_filter['sigma_kk'])
        eps_kk = np.array(group_filter['eps_kk'])
        lr_kk = np.asarray(group_filter['lr_kk'])
        la_kk = np.asarray(group_filter['la_kk'])

        Nst_kk = np.asarray(group_filter['Nst_kk'])
        sites_kk = np.array(group_filter[['nH_kk', 'ne1_kk', 'ne2_kk']])

        charge_kk = np.array(group_filter['charge_kk'])
        sigma_born_kk = np.array(group_filter['sigma_born_kk'])

        mw_kk = np.array(group_filter['mw_kk'])

        self.vki = vki
        self.subgroups = subgroups
        self.ngroups = ngroups
        self.groups_index = groups_index
        self.groups_indexes = group_indexes

        self.vk = vk
        self.Sk = Sk
        self.sigma_kk = sigma_kk
        self.eps_kk = eps_kk
        self.lr_kk = lr_kk
        self.la_kk = la_kk

        self.Nst_kk = Nst_kk
        self.sites_kk = sites_kk

        self.charge_kk = charge_kk
        self.sigma_born_kk = sigma_born_kk

        self.mw_kk = mw_kk
        self.Mw = np.dot(vki, mw_kk)

        sigma_kk3 = sigma_kk**3
        sigma_kl = np.add.outer(sigma_kk, sigma_kk)/2
        sigma_kl3 = sigma_kl**3
        la_kl = np.sqrt(np.multiply.outer(la_kk-3., la_kk-3.)) + 3
        lr_kl = np.sqrt(np.multiply.outer(lr_kk-3., lr_kk-3.)) + 3

        eps_kl = np.sqrt(np.multiply.outer(eps_kk, eps_kk))
        eps_kl *= np.sqrt(np.multiply.outer(sigma_kk3, sigma_kk3))
        eps_kl /= sigma_kl3

        for k, groupK in enumerate(subgroups):
            for l in range(k, ngtotal):
                groupL = subgroups[l]

                bool_kk = df_mie_kl.group_k == groupK
                bool_ll = df_mie_kl.group_l == groupL
                bool_kl = df_mie_kl.group_k == groupL
                bool_lk = df_mie_kl.group_l == groupK

                df1 = df_mie_kl[bool_kk & bool_ll]
                len1 = df1.shape[0]

                df2 = df_mie_kl[bool_kl & bool_lk]
                len2 = df2.shape[0]

                if len1 == 1:
                    df = df1
                    n = len1
                elif len2 == 1:
                    df = df2
                    n = len2
                else:
                    n = 0

                if n == 1:
                    _, _, eps, lr = df.values[0][0:4]

                    if eps != 'CR':
                        eps_kl[k, l] = eps
                        eps_kl[l, k] = eps
                    if lr != 'CR':
                        lr_kl[k, l] = lr
                        lr_kl[l, k] = lr

        # second order modifications
        secondorder47_1(nc, group_indexes, subgroups, vki, eps_kl,
                        df_secondorder)
        secondorder47_2(nc, group_indexes, subgroups, vki, eps_kl,
                        df_secondorder)

        asso_bool = np.any(Nst_kk > 0)
        self.asso_bool = asso_bool
        if asso_bool:
            values = asso_aux(Nst_kk, sites_kk, groups_index, subgroups,
                              df_asso_kl)
            kAB_kl, epsAB_kl, sites_asso, group_asso_index = values[0:4]
            nsites = values[4]
            DIJ = np.zeros([nsites, nsites])
            DIJ[:] = sites_asso[np.nonzero(sites_asso)]
            DIJ[epsAB_kl == 0.] = 0
            self.S = sites_asso[np.nonzero(sites_asso)]
            self.sites_asso = sites_asso
            self.kAB_kl = kAB_kl
            self.epsAB_kl = epsAB_kl
            self.DIJ = DIJ
            self.diagasso = np.diag_indices(nsites)
            self.nsites = len(sites_asso[np.nonzero(sites_asso)])
            self.group_asso_index = group_asso_index

        self.sigma_kl = sigma_kl
        self.eps_kl = eps_kl
        self.lr_kl = lr_kl
        self.la_kl = la_kl


class mixture(object):
    '''
    class mixture
    Creates an object that cointains info about a mixture.

    Parameters
    ----------
    component1 : object
        component created with component class
    component2 : object
        component created with component class

    Attributes
    ----------
    name : list
        Name of the component
    ms : list
        list of chain lenght
    sigma : list
        list of Mie diameter [m]
    eps : list
        List of Mie Potential energy [J]
    lambda_a : list
        List of Atractive exponent of Mie Potential
    lambda_r : list
        List Repulsive exponent of Mie Potential
    eAB : list
        List of Association Energy [J]
    rcAB : list
        List of Association rc [m]
    rdAB : float
        List of Association rd [m]
    sites = list
        Association sites [Bipolar, Positive, Negative]
    mupol : list
        List of dipolar moment [Debye]
    npol : list
        List of numer of polar sites de GV polar term
    Tc : list
        Critical temperature [K]
    Pc : list
        Critical Pressure [Pa]
    Zc : list
        critical compresibility factor
    Vc : list
        critical volume [m^3/mol]
    w  : list
        acentric factor
    cii : list
        polynomial coefficient for influence parameter in SGT [J m^5/mol^2]

    Methods
    -------
    add_component : adds a component to the mixture
    copy: returns a copy of the object
    kij_saft : add kij matrix for SAFT-VR-Mie
    lij_saft : add lij matrix for SAFT-VR-Mie
    ci : computes cij matrix at T for SGT
    '''

    def __init__(self, component1, component2):

        self.components = [component1, component2]
        self.names = [component1.name, component2.name]
        self.Tc = [component1.Tc, component2.Tc]
        self.Pc = [component1.Pc, component2.Pc]
        self.Zc = [component1.Zc, component2.Zc]
        self.w = [component1.w, component2.w]
        self.Vc = [component1.Vc, component2.Vc]
        self.cii = [component1.cii, component2.cii]
        self.Mw = [component1.Mw, component2.Mw]
        self.GC = [component1.GC,  component2.GC]
        self.nc = 2

        self.lr = [component1.lambda_r, component2.lambda_r]
        self.la = [component1.lambda_a, component2.lambda_a]
        self.sigma = [component1.sigma, component2.sigma]
        self.eps = [component1.eps, component2.eps]
        self.ms = [component1.ms, component2.ms]
        self.ring = [component1.ring, component2.ring]
        self.eAB = [component1.eAB, component2.eAB]
        self.rc = [component1.rcAB, component2.rcAB]
        self.rd = [component1.rdAB, component2.rdAB]
        self.sitesmix = [component1.sites, component2.sites]

        self.mupol = [component1.mupol, component2.mupol]
        self.npol = [component1.npol, component2.npol]

        ## kij matrix
        self.KIJ0saft = np.zeros([self.nc, self.nc])
        self.KIJ1saft = np.zeros([self.nc, self.nc])
        self.KIJ2saft = np.zeros([self.nc, self.nc])
        self.KIJ3saft = np.zeros([self.nc, self.nc])

        ## lij matrix
        self.LIJ0saft = np.zeros([self.nc, self.nc])
        self.LIJ1saft = np.zeros([self.nc, self.nc])
        self.LIJ2saft = np.zeros([self.nc, self.nc])
        self.LIJ3saft = np.zeros([self.nc, self.nc])

    def add_component(self, component):
        """
        add_component(component)

        Method that add a component to the mixture

        Parameters
        ----------
        component : object
            pure fluid created with component function
        """
        self.components.append(component)
        self.names.append(component.name)
        self.Tc.append(component.Tc)
        self.Pc.append(component.Pc)
        self.Zc.append(component.Zc)
        self.Vc.append(component.Vc)
        self.w.append(component.w)
        self.cii.append(component.cii)
        self.Mw.append(component.Mw)
        self.GC.append(component.GC)

        self.lr.append(component.lambda_r)
        self.la.append(component.lambda_a)
        self.sigma.append(component.sigma)
        self.eps.append(component.eps)
        self.ms.append(component.ms)
        self.ring.append(component.ring)
        self.eAB.append(component.eAB)
        self.rc.append(component.rcAB)
        self.rd.append(component.rdAB)
        self.sitesmix.append(component.sites)

        self.mupol.append(component.mupol)
        self.npol.append(component.npol)

        self.nc += 1

        # kij matrix
        self.KIJ0saft = np.zeros([self.nc, self.nc])
        self.KIJ1saft = np.zeros([self.nc, self.nc])
        self.KIJ2saft = np.zeros([self.nc, self.nc])
        self.KIJ3saft = np.zeros([self.nc, self.nc])

        # lij matrix
        self.LIJ0saft = np.zeros([self.nc, self.nc])
        self.LIJ1saft = np.zeros([self.nc, self.nc])
        self.LIJ2saft = np.zeros([self.nc, self.nc])
        self.LIJ3saft = np.zeros([self.nc, self.nc])

    def __add__(self, new_component):
        if isinstance(new_component, component):
            self.add_component(new_component)
        else:
            raise Exception('You can only add components objects to an existing mixture')
        return self

    def kij_saft(self, kij0, kij1=None, kij2=None, kij3=None):
        r"""
        kij_saft(kij)

        Method that adds kij correction for Mie potential well depth.

        .. math::
            \epsilon_{ij} = (1-k_{ij}) \frac{\sqrt{\sigma_i^3 \sigma_j^3}}{\sigma_{ij}^3} \sqrt{\epsilon_i \epsilon_j}

        kij correction is computed as follows:
        .. math::
            k_{ij} = k_{ij,0} + k_{ij,1} \cdot T +  k_{ij,2} \cdot T^2 + k_{ij,3} / T

        Parameters
        ----------
        kij0 : array_like
            kij0 matrix (Symmetric, Diagonal==0, shape=(nc, nc)) [Adim]
        kij1 : array_like, optional
            kij1 matrix (Symmetric, Diagonal==0, shape=(nc, nc)) [1/K].
            If None, then a zero matrix is assumed.
        kij2 : array_like, optional
            kij2 matrix (Symmetric, Diagonal==0, shape=(nc, nc)) [1/K^2].
            If None, then a zero matrix is assumed.
        kij3 : array_like, optional
            kij3 matrix (Symmetric, Diagonal==0, shape=(nc, nc)) [K].
            If None, then a zero matrix is assumed.

        """
        nc = self.nc
        KIJ0 = np.asarray(kij0)
        shape = KIJ0.shape

        isSquare = shape == (nc, nc)
        isSymmetric = np.allclose(KIJ0, KIJ0.T)
        diagZero = np.all(np.diagonal(KIJ0) == 0.)

        if isSquare and isSymmetric and diagZero:
            self.KIJ0saft = KIJ0
        else:
            raise Exception('kij0 matrix is not square, symmetric or diagonal==0')

        if kij1 is None:
            KIJ1 = np.zeros_like(kij0)
            self.KIJ1saft = KIJ1
        else:
            KIJ1 = np.asarray(kij1)
            shape = KIJ1.shape

            isSquare = shape == (nc, nc)
            isSymmetric = np.allclose(KIJ1, KIJ1.T)
            diagZero = np.all(np.diagonal(KIJ1) == 0.)

            if isSquare and isSymmetric and diagZero:
                self.KIJ1saft = KIJ1
            else:
                raise Exception('kij1 matrix is not square, symmetric or diagonal==0')

        if kij2 is None:
            KIJ2 = np.zeros_like(kij0)
            self.KIJ2saft = KIJ2
        else:
            KIJ2 = np.asarray(kij2)
            shape = KIJ2.shape

            isSquare = shape == (nc, nc)
            isSymmetric = np.allclose(KIJ2, KIJ2.T)
            diagZero = np.all(np.diagonal(KIJ2) == 0.)

            if isSquare and isSymmetric and diagZero:
                self.KIJ2saft = KIJ2
            else:
                raise Exception('kij2 matrix is not square, symmetric or diagonal==0')

        if kij3 is None:
            KIJ3 = np.zeros_like(kij0)
            self.KIJ3saft = KIJ3
        else:
            KIJ3 = np.asarray(kij3)
            shape = KIJ3.shape

            isSquare = shape == (nc, nc)
            isSymmetric = np.allclose(KIJ3, KIJ3.T)
            diagZero = np.all(np.diagonal(KIJ3) == 0.)

            if isSquare and isSymmetric and diagZero:
                self.KIJ3saft = KIJ3
            else:
                raise Exception('kij3 matrix is not square, symmetric or diagonal==0')

    def lij_saft(self, lij0, lij1=None, lij2=None, lij3=None):
        r"""
        lij_saft(lij)

        Method that adds lij correction for association energy.

        .. math::
            \epsilon_{ij}^{AB} = (1 - l_{ij})\sqrt{\epsilon_{ii}^{AB} \epsilon_{jj}^{AB}}

        lij correction is computed as follows:
        .. math::
            l_{ij} = l_{ij,0} + l_{ij,1} \cdot T +  l_{ij,2} \cdot T^2 + l_{ij,3} / T

        Parameters
        ----------
        lij0 : array_like
            lij0 matrix (Symmetric, Diagonal==0, shape=(nc, nc)) [Adim]
        lij1 : array_like, optional
            lij1 matrix (Symmetric, Diagonal==0, shape=(nc, nc)) [1/K].
            If None, then a zero matrix is assumed.
        lij2 : array_like, optional
            lij2 matrix (Symmetric, Diagonal==0, shape=(nc, nc)) [1/K^2].
            If None, then a zero matrix is assumed.
        lij3 : array_like, optional
            lij3 matrix (Symmetric, Diagonal==0, shape=(nc, nc)) [K]
            If None, then a zero matrix is assumed.

        """

        nc = self.nc
        LIJ0 = np.asarray(lij0)
        shape = LIJ0.shape

        isSquare = shape == (nc, nc)
        isSymmetric = np.allclose(LIJ0, LIJ0.T)
        diagZero = np.all(np.diagonal(LIJ0) == 0.)

        if isSquare and isSymmetric and diagZero:
            self.LIJ0saft = LIJ0
        else:
            raise Exception('lij0 matrix is not square, symmetric or digonal==0')

        if lij1 is None:
            LIJ1 = np.zeros_like(lij0)
            self.LIJ1saft = LIJ1
        else:
            LIJ1 = np.asarray(lij1)
            shape = LIJ1.shape

            isSquare = shape == (nc, nc)
            isSymmetric = np.allclose(LIJ1, LIJ1.T)
            diagZero = np.all(np.diagonal(LIJ1) == 0.)

            if isSquare and isSymmetric and diagZero:
                self.LIJ1saft = LIJ1
            else:
                raise Exception('lij1 matrix is not square, symmetric or diagonal==0')

        if lij2 is None:
            LIJ2 = np.zeros_like(lij0)
            self.LIJ2saft = LIJ2
        else:
            LIJ2 = np.asarray(lij2)
            shape = LIJ2.shape

            isSquare = shape == (nc, nc)
            isSymmetric = np.allclose(LIJ2, LIJ2.T)
            diagZero = np.all(np.diagonal(LIJ2) == 0.)

            if isSquare and isSymmetric and diagZero:
                self.LIJ2saft = LIJ2
            else:
                raise Exception('lij2 matrix is not square, symmetric or diagonal==0')

        if lij3 is None:
            LIJ3 = np.zeros_like(lij0)
            self.LIJ3saft = LIJ3
        else:
            LIJ3 = np.asarray(lij3)
            shape = LIJ3.shape

            isSquare = shape == (nc, nc)
            isSymmetric = np.allclose(LIJ3, LIJ3.T)
            diagZero = np.all(np.diagonal(LIJ3) == 0.)

            if isSquare and isSymmetric and diagZero:
                self.LIJ3saft = LIJ3
            else:
                raise Exception('lij3 matrix is not square, symmetric or diagonal==0')

    def set_kijsaft(self, i, j, kij0, kij1=0., kij2=0., kij3=0.):
        r"""
        set_kijsaft(i,j, kij0, kij1, kij2, kij3)

        Method that sets the kij correction for Mie potential well depth
        between component i and component j.

        .. math::
            \epsilon_{ij} = (1-k_{ij}) \frac{\sqrt{\sigma_i^3 \sigma_j^3}}{\sigma_{ij}^3} \sqrt{\epsilon_i \epsilon_j}

        kij correction is computed as follows:
        .. math::
            k_{ij} = k_{ij,0} + k_{ij,1} \cdot T +  k_{ij,2} \cdot T^2 + k_{ij,3} / T

        Parameters
        ----------
        i : int
            index of component i.
        j : int
            index of component j.
        kij0 : float
            kij0 value between component i and j [Adim]
        kij1 : float, optional
            kij1 value between component i and j [1/K]. Default to zero.
        kij2 : float, optional
            kij2 value between component i and j [1/K^2]. Default to zero.
        kij3 : float, optional
            kij3 value between component i and j [K]. Default to zero.

        """

        typei = type(i) == int
        typej = type(j) == int

        nc = self.nc
        nc_i = 0 <= i <= (nc - 1)
        nc_j = 0 <= j <= (nc - 1)

        i_j = i != j

        if (not nc_i) or (not nc_j):
            raise Exception('Index i or j bigger than (nc-1)')
        if not i_j:
            raise Exception('Cannot set kij for i=j')

        if typei and typej and nc_i and nc_j and i_j:
            self.KIJ0saft[i, j] = kij0
            self.KIJ0saft[j, i] = kij0

            self.KIJ1saft[i, j] = kij1
            self.KIJ1saft[j, i] = kij1

            self.KIJ2saft[i, j] = kij2
            self.KIJ2saft[j, i] = kij2

            self.KIJ3saft[i, j] = kij3
            self.KIJ3saft[j, i] = kij3

    def set_lijsaft(self, i, j, lij0, lij1=0., lij2=0., lij3=0.):
        r"""
        set_lijsaft(i,j, lij0, lij1, lij2, lij3)

        Method that sets the lij correction for cross-association energy
        between component i and component j.

        .. math::
            \epsilon_{ij}^{AB} = (1 - l_{ij})\sqrt{\epsilon_{ii}^{AB} \epsilon_{jj}^{AB}}

        lij correction is computed as follows:
        .. math::
            l_{ij} = l_{ij,0} + l_{ij,1} \cdot T +  l_{ij,2} \cdot T^2 + l_{ij,3} / T

        Parameters
        ----------
        i : int
            index of component i.
        j : int
            index of component j.
        lij0 : float
            lij0 value between component i and j [Adim]
        lij1 : float, optional
            lij1 value between component i and j [1/K]. Default to zero.
        lij2 : float, optional
            lij2 value between component i and j [1/K^2]. Default to zero.
        lij3 : float, optional
            lij3 value between component i and j [K]. Default to zero.

        """

        typei = type(i) == int
        typej = type(j) == int

        nc = self.nc
        nc_i = 0 <= i <= (nc - 1)
        nc_j = 0 <= j <= (nc - 1)

        i_j = i != j

        if (not nc_i) or (not nc_j):
            raise Exception('Index i or j bigger than (nc-1)')
        if not i_j:
            raise Exception('Cannot set lij for i=j')

        if typei and typej and nc_i and nc_j and i_j:
            self.LIJ0saft[i, j] = lij0
            self.LIJ0saft[j, i] = lij0

            self.LIJ1saft[i, j] = lij1
            self.LIJ1saft[j, i] = lij1

            self.LIJ2saft[i, j] = lij2
            self.LIJ2saft[j, i] = lij2

            self.LIJ3saft[i, j] = lij3
            self.LIJ3saft[j, i] = lij3

    def ci(self, T):
        """
        Method that computes the matrix of cij interaction parameter for SGT at
        given temperature.

        Parameters
        ----------
        T : float
            absolute temperature [K]

        Returns
        -------
        ci : array_like
            influence parameter matrix at given temperature [J m^5 / mol^2]

        """

        n = len(self.cii)
        ci = np.zeros(n)
        for i in range(n):
            ci[i] = np.polyval(self.cii[i], T)
        self.cij = np.sqrt(np.outer(ci, ci))
        return self.cij

    def copy(self):
        """
        Method that return a copy of the mixture

        Returns
        -------
        mix : object
            returns a copy a of the mixture
        """
        return copy(self)

    def saftgammamie(self, database=database):
        """
        saftgammamie method

        Method that reads the SAFT-Gamma-Mie database and set the necessary
        interactions parameters to use the equation of state.

        Parameters
        ----------
        database: Object
            database object
        """
        df_groups = database.df_groups
        df_mie_kl = database.df_mie_kl
        df_asso_kl = database.df_asso_kl
        df_secondorder = database.df_secondorder
        df_secondasso = database.df_secondasso

        GC = self.GC
        nc = self.nc

        subgroups = []
        ngroups = []
        vki = []
        for i in range(nc):
            group_names = list(GC[i].keys())
            group_vki = list(GC[i].values())
            subgroups += group_names
            vki += group_vki
            ngroups.append(len(group_names))

        subgroups = np.asarray(subgroups)
        vki = np.asarray(vki)

        ngtotal = np.sum(ngroups)
        groups_index = np.zeros(ngtotal, dtype=np.int64)
        group_indexes = []

        index0 = 0
        for i in range(nc):
            indexf = ngroups[i] + index0
            group_indexes.append([index0, indexf])
            groups_index[index0:indexf] = i
            index0 = indexf

        group_filter = df_groups.loc[subgroups]

        vk = np.asarray(group_filter['vk*'])
        Sk = np.asarray(group_filter['Sk'])

        sigma_kk = np.asarray(group_filter['sigma_kk'])
        eps_kk = np.array(group_filter['eps_kk'])
        lr_kk = np.asarray(group_filter['lr_kk'])
        la_kk = np.asarray(group_filter['la_kk'])

        Nst_kk = np.asarray(group_filter['Nst_kk'])
        sites_kk = np.array(group_filter[['nH_kk', 'ne1_kk', 'ne2_kk']])

        charge_kk = np.array(group_filter['charge_kk'])
        sigma_born_kk = np.array(group_filter['sigma_born_kk'])

        mw_kk = np.array(group_filter['mw_kk'])

        sigma_kk3 = sigma_kk**3
        sigma_kl = np.add.outer(sigma_kk, sigma_kk)/2
        sigma_kl3 = sigma_kl**3
        la_kl = np.sqrt(np.multiply.outer(la_kk-3., la_kk-3.)) + 3
        lr_kl = np.sqrt(np.multiply.outer(lr_kk-3., lr_kk-3.)) + 3

        eps_kl = np.sqrt(np.multiply.outer(eps_kk, eps_kk))
        eps_kl *= np.sqrt(np.multiply.outer(sigma_kk3, sigma_kk3))
        eps_kl /= sigma_kl3

        for k, groupK in enumerate(subgroups):
            for l in range(k, ngtotal):
                groupL = subgroups[l]

                bool_kk = df_mie_kl.group_k == groupK
                bool_ll = df_mie_kl.group_l == groupL
                bool_kl = df_mie_kl.group_k == groupL
                bool_lk = df_mie_kl.group_l == groupK

                df1 = df_mie_kl[bool_kk & bool_ll]
                len1 = df1.shape[0]

                df2 = df_mie_kl[bool_kl & bool_lk]
                len2 = df2.shape[0]

                if len1 == 1:
                    df = df1
                    n = len1
                elif len2 == 1:
                    df = df2
                    n = len2
                else:
                    n = 0

                if n == 1:
                    _, _, eps, lr = df.values[0][0:4]

                    if eps != 'CR':
                        eps_kl[k, l] = eps
                        eps_kl[l, k] = eps
                    if lr != 'CR':
                        lr_kl[k, l] = lr
                        lr_kl[l, k] = lr

        self.vki = vki
        self.subgroups = subgroups
        self.ngroups = ngroups
        self.ngtotal = ngtotal
        self.groups_index = groups_index
        self.groups_indexes = group_indexes

        self.vk = vk
        self.Sk = Sk
        self.sigma_kk = sigma_kk
        self.eps_kk = eps_kk
        self.lr_kk = lr_kk
        self.la_kk = la_kk

        self.Nst_kk = Nst_kk
        self.sites_kk = sites_kk

        self.charge_kk = charge_kk
        self.sigma_born_kk = sigma_born_kk

        self.mw_kk = mw_kk

        Mw_i = np.zeros(nc)
        for i in range(nc):
            i0 = group_indexes[i][0]
            i1 = group_indexes[i][1]
            Mw_i[i] = np.dot(vki[i0:i1], mw_kk[i0:i1])
        self.Mw = Mw_i

        # second order modifications
        secondorder47_1(nc, group_indexes, subgroups, vki, eps_kl,
                        df_secondorder)
        secondorder47_2(nc, group_indexes, subgroups, vki, eps_kl,
                        df_secondorder)

        asso_bool = np.any(Nst_kk > 0)
        self.asso_bool = asso_bool
        if asso_bool:
            values = asso_aux(Nst_kk, sites_kk, groups_index, subgroups,
                              df_asso_kl)
            kAB_kl, epsAB_kl, sites_asso, group_asso_index = values[0:4]
            nsites, molecule_id_index_sites = values[4:6]
            indexAB_id, indexABij, subgroup_id_asso = values[6:9]
            molecule_id_index_asso, sites_cumsum = values[9:11]

            secondorder19_14(nc, group_indexes, subgroups, vki, eps_kl, lr_kl,
                             df_secondorder, subgroup_id_asso,
                             molecule_id_index_asso, sites_cumsum, epsAB_kl,
                             kAB_kl, df_secondasso)

            secondorder21(nc, group_indexes, subgroups, vki, eps_kl, lr_kl,
                          df_secondorder, subgroup_id_asso,
                          molecule_id_index_asso, sites_cumsum, epsAB_kl,
                          kAB_kl, df_secondasso)

            secondorder22(nc, group_indexes, subgroups, vki, eps_kl, lr_kl,
                          df_secondorder, subgroup_id_asso,
                          molecule_id_index_asso, sites_cumsum, epsAB_kl,
                          kAB_kl, df_secondasso)

            secondorder23(nc, group_indexes, subgroups, vki, eps_kl, lr_kl,
                          df_secondorder, subgroup_id_asso,
                          molecule_id_index_asso, sites_cumsum, epsAB_kl,
                          kAB_kl, df_secondasso)

            DIJ = np.zeros([nsites, nsites])
            DIJ[:] = sites_asso[np.nonzero(sites_asso)]
            DIJ[epsAB_kl == 0.] = 0
            self.S = sites_asso[np.nonzero(sites_asso)]
            self.sites_asso = sites_asso
            self.kAB_kl = kAB_kl
            self.epsAB_kl = epsAB_kl
            self.DIJ = DIJ
            self.diagasso = np.diag_indices(nsites)
            self.nsites = nsites
            self.group_asso_index = group_asso_index
            self.molecule_id_index_sites = molecule_id_index_sites
            self.indexAB_id = indexAB_id
            self.indexABij = indexABij
            self.vki_asso = vki[self.group_asso_index]

            dxjdx = np.zeros([nc, nsites])
            for i in range(nc):
                dxjdx[i] = molecule_id_index_sites == i
                dxjdx[i] *= self.vki_asso
            self.dxjasso_dx = dxjdx

        self.sigma_kl = sigma_kl
        self.eps_kl = eps_kl
        self.lr_kl = lr_kl
        self.la_kl = la_kl



from __future__ import division, print_function, absolute_import
import numpy as np


# Equation 28
def I_lam(x0, lam):
    lam3 = 3. - lam
    I = (x0**lam3 - 1.) / lam3
    return I

# Equation 29
def J_lam(x0, lam):
    lam3 = 3. - lam
    lam4 = 4. - lam
    J = (lam3*x0**lam4 - lam4*x0**lam3 + 1.)/(lam3 * lam4)
    return J


# Equation 16
def kHS(eta):
    num = (1-eta)**4
    den = 1 + 4*eta+4*eta**2-4*eta**3 + eta**4
    return num/den


def dkHS(eta):
    den = 1 + 4*eta+4*eta**2-4*eta**3 + eta**4
    num1 = (1-eta)**4
    num2 = -4*(-1 + eta)**3*(-2 + (-5 + eta)*eta)

    khs = num1/den
    dkhs = num2/den**2

    return khs, dkhs


def d2kHS(eta):
    den = 1 + 4*eta+4*eta**2-4*eta**3 + eta**4
    num1 = (1-eta)**4
    num2 = -4*(-1 + eta)**3*(-2 + (-5 + eta)*eta)

    num3 = 4*(-1 + eta)**2*(17 + eta*(82 + eta*(39 +
           eta*(-80 + eta*(77 + 3*(-10 + eta)*eta)))))
    khs = num1/den
    dkhs = num2/den**2
    d2khs = num3/den**3
    return khs, dkhs, d2khs


def d3kHS(eta):
    den = 1 + 4*eta+4*eta**2-4*eta**3 + eta**4

    num1 = (1-eta)**4
    num2 = -4*(-1 + eta)**3*(-2 + (-5 + eta)*eta)

    num3 = 4*(-1 + eta)**2*(17 + eta*(82 + eta*(39 +
           eta*(-80 + eta*(77 + 3*(-10 + eta)*eta)))))

    num4 = -624. - 4032. * eta - 576. * eta**2 + 16656. * eta**3-11424.*eta**4
    num4 += -16896. * eta**5 + 34752. * eta**6 - 27936. * eta**7+13776.*eta**8
    num4 += - 4416. * eta**9 + 768. * eta**10 - 48. * eta**11
    khs = num1/den
    dkhs = num2/den**2
    d2khs = num3/den**3
    d3khs = num4/den**4
    return khs, dkhs, d2khs, d3khs


def eta_eff(eta, ci):
    eta_vec = np.array([eta, eta**2, eta**3, eta**4])
    neff = np.dot(ci, eta_vec)
    return neff


def deta_eff(eta, ci):

    eta_vec = np.array([[eta, eta**2, eta**3, eta**4],
                       [1., 2*eta, 3*eta**2, 4*eta**3]])

    dneff = np.matmul(eta_vec, ci)
    return dneff


def d2eta_eff(eta, ci):

    eta_vec = np.array([[eta, eta**2, eta**3, eta**4],
                       [1., 2*eta, 3*eta**2, 4*eta**3],
                       [0., 2., 6.*eta, 12.*eta**2]])
    d2neff = np.matmul(eta_vec, ci)

    return d2neff


def d3eta_eff(eta, ci):
    eta_vec = np.array([[eta, eta**2, eta**3, eta**4],
                       [1., 2*eta, 3*eta**2, 4*eta**3],
                       [0., 2., 6.*eta, 12.*eta**2],
                       [0., 0., 6., 24.*eta]])

    d3neff = np.matmul(eta_vec, ci)

    return d3neff


# Equation 17
def Xi(x03, nsigma, f1, f2, f3):
    x = f1*nsigma + f2*nsigma**5 + f3*nsigma**8
    return x


def dXi(x03, nsigma, f1, f2, f3):
    x = f1*nsigma + f2*nsigma**5 + f3*nsigma**8
    dx = x03*(f1 + 5*f2 * nsigma**4 + 8*f3*nsigma**7)
    return x, dx


def d2Xi(x03, nsigma, f1, f2, f3):
    x = f1*nsigma + f2*nsigma**5 + f3*nsigma**8
    dx = x03*(f1 + 5*f2 * nsigma**4 + 8*f3*nsigma**7)
    d2x = x03**2*(20.*f2 * nsigma**3 + 56.*f3*nsigma**6)
    return x, dx, d2x



from __future__ import division, print_function, absolute_import
import numpy as np
from .stability import tpd_minimas
from .multiflash import multiflash


def lle(x0, w0, Z, T, P, model, v0=[None, None], Xass0=[None, None],
        K_tol=1e-8, nacc=5, full_output=False, tetha_max=10.):
    """
    Liquid liquid equilibrium (z,T,P) -> (x,w,beta)

    Solves liquid liquid equilibrium from multicomponent mixtures at given
    pressure, temperature and overall composition.

    Parameters
    ----------

    x0 : array_like
        initial guess for liquid phase 1
    w0 : array_like
        initial guess for liquid phase 2
    z : array_like
        overal composition of mix
    T : float
        absolute temperature [K].
    P : float
        pressure [Pa]
    model : object
        created from mixture and saftvrmie function
    v0 : list, optional
        if supplied volume used as initial value to compute fugacities
    K_tol : float, optional
        Desired accuracy of K (= W/X) vector
    nacc : int, optional
        number of accelerated successive substitution cycles to perform
    full_output: bool, optional
        wheter to outputs all calculation info
    tetha_max : float, optional
        maximum value for stability variables. Default is 10.

    Returns
    -------
    X : array_like
        liquid 1 mole fraction vector
    W : array_like
        liquid 2 mole fraction vector
    beta : float
        phase fraction of liquid 2

    """
    nc = model.nc
    if len(x0) != nc or len(w0) != nc or len(Z) != nc:
        raise Exception('Composition vector lenght must be equal to nc')

    temp_aux = model.temperature_aux(T)

    equilibrio = ['L', 'L']
    fugx, v1, Xass1 = model.logfugef_aux(x0, temp_aux, P, equilibrio[0], v0[0],
                                         Xass0[0])
    fugw, v2, Xass2 = model.logfugef_aux(w0, temp_aux, P, equilibrio[1], v0[1],
                                         Xass0[1])
    lnK = fugx - fugw
    K = np.exp(lnK)

    bmin = max(np.hstack([((K*Z-1.)/(K-1.))[K > 1], 0.]))
    bmax = min(np.hstack([((1.-Z)/(1.-K))[K < 1], 1.]))
    beta = (bmin + bmax)/2
    X0 = np.array([x0, w0])

    beta0 = np.array([1-beta, beta, 0.])

    out = multiflash(X0, beta0, equilibrio, Z, T, P, model,
                     [v1, v2], [Xass1, Xass2], K_tol, nacc, True, tetha_max)
    Xm, beta, tetha, v, Xass = out.X, out.beta, out.tetha, out.v, out.Xass

    if tetha > 0:
        xes, tpd_min2 = tpd_minimas(2, Xm[0], T, P, model, 'L', 'L',
                                    v[0], v[0])
        X0 = np.asarray(xes)
        beta0 = np.hstack([beta, 0.])
        out = multiflash(X0, beta0, equilibrio, Z, T, P, model, v, Xass,
                         K_tol, nacc, True, tetha_max)
        Xm, beta, tetha, v = out.X, out.beta, out.tetha, out.v

    X, W = Xm
    if tetha > 0:
        W = X.copy()

    if full_output:
        return out

    return X, W, beta[1]


__all__ = ['lle']



from __future__ import division, print_function, absolute_import
import numpy as np
from .tensionresult import TensionResult
from ..math import gauss, colocAB
from scipy.optimize import root
from scipy.interpolate import interp1d
from .cijmix_cy import cmix_cy


def fobj_z_newton(rointer, Binter, dro20, dro21, mu0, temp_aux, cij, n, nc,
                  model, Xass01):
    rointer = np.abs(rointer.reshape([nc, n]))
    dmu = np.zeros([n, nc])

    i = 0
    dmu[i], Xass = model.muad_aux(rointer[:, i], temp_aux, Xass01)

    for i in range(1, n):
        dmu[i], Xass = model.muad_aux(rointer[:, i], temp_aux, Xass)
    dmu -= mu0
    dmu = dmu.T

    dro2 = np.matmul(rointer, Binter.T)
    dro2 += dro20
    dro2 += dro21

    ter1 = np.matmul(cij, dro2)
    fo = ter1 - dmu
    return fo.flatten()


def dfobj_z_newton(rointer, Binter, dro20, dro21, mu0, temp_aux, cij, n, nc,
                   model, Xass01):

    index0 = rointer < 0
    rointer = np.abs(rointer.reshape([nc, n]))
    dmu = np.zeros([n, nc])
    d2mu = np.zeros([n, nc, nc])

    i = 0
    dmu[i], d2mu[i], Xass = model.dmuad_aux(rointer[:, i], temp_aux, Xass01)
    for i in range(n):
        dmu[i], d2mu[i], Xass = model.dmuad_aux(rointer[:, i], temp_aux, Xass)
    dmu -= mu0
    dmu = dmu.T
    d2mu = d2mu.T

    dro2 = np.matmul(rointer, Binter.T)
    dro2 += dro20
    dro2 += dro21

    ter1 = np.matmul(cij, dro2)
    fo = ter1 - dmu

    eye = np.eye(n, n)
    dml = []
    for i in range(nc):
        auxlist = []
        for j in range(nc):
            auxlist.append(eye * d2mu[i, j])
        dml.append(auxlist)
    jac_dmu = np.block(dml)

    jacter1 = np.multiply.outer(cij, Binter.T).reshape(nc, nc*n, n)
    fo_jac = np.hstack(jacter1).T - jac_dmu
    fo_jac[:, index0] *= -1

    return fo.flatten(), fo_jac


def sgt_mix(rho1, rho2, Tsat, Psat, model, rho0='linear',
            z0=10., dz=1.5, itmax=10, n=20, full_output=False,
            ten_tol=1e-2, root_method='lm', solver_opt=None,
            check_eq=True):
    """
    SGT for mixtures and beta != 0 (rho1, rho2, T, P) -> interfacial tension

    Parameters
    ----------
    rho1 : float
        phase 1 density vector
    rho2 : float
        phase 2 density vector
    Tsat : float
        saturation temperature
    Psat : float
        saturation pressure
    model : object
        created with an EoS
    rho0 : string, array_like or TensionResult
        inital values to solve the BVP, avaialable options are 'linear' for
        linear density profiles, 'hyperbolic' for hyperbolic
        like density profiles. An array can also be supplied or a
        TensionResult of a previous calculation.
    z0 :  float
        initial interfacial lenght
    dz : float
        increase in interfacial lenght per interation
    itmax : int
        maximun number of iterations
    n : int, optional
        number points to solve density profiles
    full_output : bool, optional
        wheter to outputs all calculation info
    root_method: string, optional
        Method used un SciPy's root function
        default 'lm',  other options: 'krylov', 'hybr'. See SciPy documentation
        for more info
    solver_opt : dict, optional
        aditional solver options passed to SciPy solver
    check_eq : bool, optional
        whether to check if given density vectors are in phase equilibria

    Returns
    -------
    ten : float
        interfacial tension between the phases
    """

    z = z0
    nc = model.nc

    # Dimensionless Variables
    Tfactor, Pfactor, rofactor, tenfactor, zfactor = model.sgt_adim(Tsat)
    Pad = Psat*Pfactor
    rho1a = rho1*rofactor
    rho2a = rho2*rofactor

    cij = model.ci(Tsat)
    cij /= cij[0, 0]
    dcij = np.linalg.det(cij)
    if np.isclose(dcij, 0):
        warning = 'Determinant of influence parameters matrix is: {}'
        raise Exception(warning.format(dcij))

    temp_aux = model.temperature_aux(Tsat)

    # Chemical potential
    mu0, Xass01 = model.muad_aux(rho1a, temp_aux)
    mu02, Xass02 = model.muad_aux(rho2a, temp_aux)
    if check_eq:
        if not np.allclose(mu0, mu02, rtol=1e-3):
            raise Exception('Not equilibria compositions, mu1 != mu2')

    # Nodes and weights of integration
    roots, weights = gauss(n)
    rootsf = np.hstack([0., roots, 1.])
    # Coefficent matrix for derivatives
    A, B = colocAB(rootsf)

    # Initial profiles
    if isinstance(rho0, str):
        if rho0 == 'linear':
            # Linear Profile
            pend = (rho2a - rho1a)
            b = rho1a
            pfl = (np.outer(roots, pend) + b)
            rointer = (pfl.T).copy()
        elif rho0 == 'hyperbolic':
            # Hyperbolic profile
            inter = 8*roots - 4
            thb = np.tanh(2*inter)
            pft = np.outer(thb, (rho2a-rho1a))/2+(rho1a+rho2a)/2
            rointer = pft.T
        else:
            raise Exception('Initial density profile not known')
    elif isinstance(rho0,  TensionResult):
        _z0 = rho0.z
        _ro0 = rho0.rho
        z = _z0[-1]
        rointer = interp1d(_z0, _ro0)(roots * z)
        rointer *= rofactor
    elif isinstance(rho0,  np.ndarray):
        # Check dimensiones
        if rho0.shape[0] == nc and rho0.shape[1] == n:
            rointer = rho0.copy()
            rointer *= rofactor
        else:
            raise Exception('Shape of initial value must be nc x n')
    else:
        raise Exception('Initial density profile not known')

    if root_method == 'lm' or root_method == 'hybr':
        if model.secondordersgt:
            fobj = dfobj_z_newton
            jac = True
        else:
            fobj = fobj_z_newton
            jac = None
    else:
        fobj = fobj_z_newton
        jac = None

    error = 1.
    it = 0
    ten_old = 0.
    tol = ten_tol
    while error > tol and it < itmax:
        it += 1
        zad = z*zfactor
        Ar = A/zad
        Br = B/zad**2

        Binter = Br[1:-1, 1:-1]
        B0 = Br[1:-1, 0]
        B1 = Br[1:-1, -1]
        dro20 = np.outer(rho1a, B0)  # cte
        dro21 = np.outer(rho2a, B1)  # cte

        Ainter = Ar[1:-1, 1:-1]
        A0 = Ar[1:-1, 0]
        A1 = Ar[1:-1, -1]
        dro10 = np.outer(rho1a, A0)  # cte
        dro11 = np.outer(rho2a, A1)  # cte

        sol = root(fobj, rointer.flatten(), method=root_method, jac=jac,
                   args=(Binter, dro20, dro21, mu0, temp_aux, cij, n, nc,
                   model, Xass01), options=solver_opt)

        rointer = sol.x
        rointer = np.abs(rointer.reshape([nc, n]))

        dro = np.matmul(rointer, Ainter.T)
        dro += dro10
        dro += dro11

        suma = cmix_cy(dro, cij)
        dom = np.zeros(n)
        k = 0
        dom[k], Xass = model.dOm_aux(rointer[:, k], temp_aux, mu0, Pad, Xass01)
        for k in range(1, n):
            dom[k], Xass = model.dOm_aux(rointer[:, k], temp_aux, mu0, Pad,
                                         Xass)
        dom[dom < 0] = 0.
        intten = np.nan_to_num(np.sqrt(2*suma*dom))
        ten = np.dot(intten, weights)
        ten *= zad
        ten *= tenfactor
        error = np.abs(ten - ten_old)
        ten_old = ten
        z += dz

    fun_error = np.linalg.norm(sol.fun)/n/nc
    success = sol.success and (error < tol)

    if full_output:

        znodes = (z-dz) * rootsf
        ro = np.insert(rointer, 0, rho1a, axis=1)
        ro = np.insert(ro, n+1, rho2a, axis=1)
        ro /= rofactor
        dictresult = {'tension': ten, 'rho': ro, 'z': znodes,
                      'GPT': np.hstack([0, dom, 0]),
                      'success': success,
                      'message': sol.message,
                      'fun_norm': fun_error,
                      'error': error, 'iter': it}
        out = TensionResult(dictresult)
        return out

    return ten


# Function for solving sgt for a fixed interfacial lenght
def sgt_zfixed(rho1, rho2, Tsat, Psat, model, rho0='linear',
               z=10, n=20, full_output=False, solver_opt=None,
               check_eq=True):
    """
    SGT for mixtures and beta != 0 (rho1, rho2, T, P) -> interfacial tension

    Parameters
    ----------
    rho1 : float
        phase 1 density vector
    rho2 : float
        phase 2 density vector
    Tsat : float
        saturation temperature
    Psat : float
        saturation pressure
    model : object
        created with an EoS
    rho0 : string, array_like or TensionResult
        inital values to solve the BVP, avaialable options are 'linear' for
        linear density profiles, 'hyperbolic' for hyperbolic
        like density profiles. An array can also be supplied or a
        TensionResult of a previous calculation.
    z :  float
        initial interfacial lenght
    n : int, optional
        number points to solve density profiles
    full_output : bool, optional
        wheter to outputs all calculation info
    solver_opt : dict, optional
        aditional solver options passed to SciPy solver
    check_eq : bool, optional
        whether to check if given density vectors are in phase equilibria

    Returns
    -------
    ten : float
        interfacial tension between the phases
    """

    nc = model.nc

    # Dimensionless Variables
    Tfactor, Pfactor, rofactor, tenfactor, zfactor = model.sgt_adim(Tsat)
    Pad = Psat*Pfactor
    rho1a = rho1*rofactor
    rho2a = rho2*rofactor

    cij = model.ci(Tsat)
    cij /= cij[0, 0]
    dcij = np.linalg.det(cij)
    if np.isclose(dcij, 0):
        warning = 'Determinant of influence parameters matrix is: {}'
        raise Exception(warning.format(dcij))

    temp_aux = model.temperature_aux(Tsat)

    # Chemical potential
    mu0, Xass01 = model.muad_aux(rho1a, temp_aux)
    mu02, Xass02 = model.muad_aux(rho2a, temp_aux)
    if check_eq:
        if not np.allclose(mu0, mu02, rtol=1e-3):
            raise Exception('Not equilibria compositions, mu1 != mu2')

    # Nodes and weights of integration
    roots, weights = gauss(n)
    rootsf = np.hstack([0., roots, 1.])
    # Coefficent matrix for derivatives
    A, B = colocAB(rootsf)

    # Initial profiles
    if rho0 == 'linear':
        # Linear Profile
        pend = (rho2a - rho1a)
        b = rho1a
        pfl = (np.outer(roots, pend) + b)
        rointer = (pfl.T).copy()
    elif rho0 == 'hyperbolic':
        # Hyperbolic profile
        inter = 8*roots - 4
        thb = np.tanh(2*inter)
        pft = np.outer(thb, (rho2a-rho1a))/2+(rho1a+rho2a)/2
        rointer = pft.T
    elif isinstance(rho0,  TensionResult):
        _z0 = rho0.z
        _ro0 = rho0.rho
        z = _z0[-1]
        rointer = interp1d(_z0, _ro0)(roots * z)
        rointer *= rofactor
    elif isinstance(rho0,  np.ndarray):
        # Check dimensiones
        if rho0.shape[0] == nc and rho0.shape[1] == n:
            rointer = rho0.copy()
            rointer *= rofactor
        else:
            raise Exception('Shape of initial value must be nc x n')

    zad = z*zfactor
    Ar = A/zad
    Br = B/zad**2

    Binter = Br[1:-1, 1:-1]
    B0 = Br[1:-1, 0]
    B1 = Br[1:-1, -1]
    dro20 = np.outer(rho1a, B0)  # cte
    dro21 = np.outer(rho2a, B1)  # cte

    Ainter = Ar[1:-1, 1:-1]
    A0 = Ar[1:-1, 0]
    A1 = Ar[1:-1, -1]
    dro10 = np.outer(rho1a, A0)  # cte
    dro11 = np.outer(rho2a, A1)  # cte

    if model.secondordersgt:
        fobj = dfobj_z_newton
        jac = True
    else:
        fobj = fobj_z_newton
        jac = None

    sol = root(fobj, rointer.flatten(), method='lm', jac=jac,
               args=(Binter, dro20, dro21, mu0, temp_aux, cij, n, nc, model,
               Xass01), options=solver_opt)

    rointer = sol.x
    rointer = np.abs(rointer.reshape([nc, n]))

    dro = np.matmul(rointer, Ainter.T)
    dro += dro10
    dro += dro11

    suma = cmix_cy(dro, cij)
    dom = np.zeros(n)
    k = 0
    dom[k], Xass = model.dOm_aux(rointer[:, k], temp_aux, mu0, Pad, Xass01)
    for k in range(1, n):
        dom[k], Xass = model.dOm_aux(rointer[:, k], temp_aux, mu0, Pad, Xass)
    dom[dom < 0] = 0.
    intten = np.nan_to_num(np.sqrt(2*suma*dom))
    ten = np.dot(intten, weights)
    ten *= zad
    ten *= tenfactor
    fun_error = np.linalg.norm(sol.fun)/n/nc
    success = sol.success

    if full_output:

        znodes = z * rootsf
        ro = np.insert(rointer, 0, rho1a, axis=1)
        ro = np.insert(ro, n+1, rho2a, axis=1)
        ro /= rofactor
        dictresult = {'tension': ten, 'rho': ro, 'z': znodes,
                      'GPT': np.hstack([0, dom, 0]),
                      'success': success,
                      'message': sol.message,
                      'fun_norm': fun_error}
        out = TensionResult(dictresult)
        return out

    return ten



from __future__ import division, print_function, absolute_import
import numpy as np


def a1(xs, a1kl):
    a = np.matmul(np.matmul(a1kl, xs), xs)
    return a


def da1_dxhi00(xs, da1kl):
    da = np.matmul(np.matmul(da1kl, xs), xs)
    return da


def d2a1_dxhi00(xs, d2a1kl):
    d2a = np.matmul(np.matmul(d2a1kl, xs), xs)
    return d2a


def d3a1_dxhi00(xs, d3a1kl):
    d3a = np.matmul(np.matmul(d3a1kl, xs), xs)
    return d3a


def da1_dx(xs, dxs_dx, a1kl, da1kl_dx):

    a = np.matmul(np.matmul(a1kl, xs), xs)

    aux1 = xs * a1kl
    suma1 = 2*np.sum(dxs_dx@aux1, axis=1)

    da1x = np.matmul(np.matmul(da1kl_dx, xs), xs)
    da1x += suma1

    return a, da1x


def da1_dxxhi(xs, dxs_dx, da1kl, da1kl_dx):

    a = np.matmul(np.matmul(da1kl, xs), xs)

    aux1 = xs * da1kl[0]
    suma1 = 2*np.sum(dxs_dx@aux1, axis=1)

    da1x = np.matmul(np.matmul(da1kl_dx, xs), xs)
    da1x += suma1

    return a, da1x



from __future__ import division, print_function, absolute_import
import numpy as np
from scipy.optimize import minimize
from ..math import gdem
from .equilibriumresult import EquilibriumResult


def multiflash_obj(inc, Z, K):

    inc[inc <= 1e-10] = 1e-10
    beta, tetha = np.array_split(inc, 2)
    ninc = len(inc)
    nbeta = len(beta)

    betar = beta[1:]
    Kexp = K.T*np.exp(tetha)
    K1 = Kexp - 1.  # nc x P - 1
    sum1 = 1 + K1@betar  # P - 1
    Xref = Z/sum1

    f1 = beta.sum() - 1.
    f2 = (K1.T*Xref).sum(axis=1)
    betatetha = (betar + tetha)
    f3 = (betar*tetha) / betatetha
    f = np.hstack([f1, f2, f3])

    jac = np.zeros([ninc, ninc])
    Id = np.eye(nbeta-1)

    # f1 derivative
    f1db = np.ones(nbeta)
    jac[0, :nbeta] = f1db

    # f2 derivative
    ter1 = Xref/sum1 * K1.T
    dfb = - ter1@K1
    jac[1:nbeta, 1:nbeta] = dfb
    dft = Id * (Xref*Kexp.T).sum(axis=1)
    ter2 = Kexp*betar
    dft -= ter1 @ ter2
    jac[1:nbeta, nbeta:] = dft

    # f3 derivative
    f3db = (tetha/betatetha)**2
    f3dt = (betar/betatetha)**2
    jac[nbeta:, 1:nbeta] = Id*f3db
    jac[nbeta:, nbeta:] = Id*f3dt

    return f, jac, Kexp, Xref


def gibbs_obj(ind, phases, Z, z_notzero, n_nonzero, temp_aux, P, model):

    nfase = len(phases)

    nc = model.nc
    ind = 1.*ind.reshape(nfase-1, n_nonzero)
    dep = Z[z_notzero] - ind.sum(axis=0)

    X = np.zeros((nfase, nc))
    X[1:, z_notzero] = ind
    X[0, z_notzero] = dep

    n = X.copy()
    X = (X.T/X.sum(axis=1)).T

    lnphi = np.zeros_like(X)
    global vg, Xassg

    for i, state in enumerate(phases):
        lnphi[i], vg[i], Xassg[i] = model.logfugef_aux(X[i], temp_aux, P,
                                                       state, vg[i], Xassg[i])
        if np.any(np.isnan(lnphi[i])) or np.isnan(vg[i]):
            lnphi[i], vg[i], Xassg[i] = model.logfugef_aux(X[i], temp_aux, P, state)

    with np.errstate(invalid='ignore'):
        fug = np.nan_to_num(np.log(X[:, z_notzero]) + lnphi[:, z_notzero])
    G = np.sum(n[:, z_notzero] * fug)
    dG = (fug[1:] - fug[0]).flatten()
    return G, dG


def multiflash(X0, betatetha, equilibrium, z, T, P, model, v0=[None],
               Xass0=[None], K_tol=1e-10, nacc=5, full_output=False,
               tetha_max=10.):
    """
    multiflash (z,T,P) -> (x, w, y, ..., beta)

    Parameters
    ----------
    X0 : array_like
        (n_phases, nc) initial guess
    betatheta : array_like
        phase fractions and stability array
    equilibrium : list
        'L' for liquid, 'V' for vaopur phase. ['L', 'L', 'V'] for LLVE
    z : array_like
        overall system composition
    T : float
        absolute temperature [K].
    P : float
        pressure [Pa]
    model : object
        created from mixture and saftvrmie function
    v0 : list, optional
        if supplied volume used as initial value to compute fugacities
    K_tol : float, optional
        Desired accuracy of K (= X/Xr) vector
    nacc : int, optional
        number of accelerated successive substitution cycles to perform
    full_output: bool, optional
        wheter to outputs all calculation info
    tetha_max : float, optional
        maximum value for stability variables. Default is 10.

    Returns
    -------
    X : array_like
        composition values matrix
    beta : array_like
        phase fraction array
    theta : array_like
        stability variables arrays

    """

    nfase = len(equilibrium)

    nc = model.nc
    if np.all(X0.shape != (nfase, nc)) or len(z) != nc:
        raise Exception('Composition vector lenght must be equal to nc')

    error = 1
    it = 0
    itacc = 0
    ittotal = 0
    n = 5

    temp_aux = model.temperature_aux(T)

    X = X0.copy()
    lnphi = np.zeros_like(X)

    if len(v0) != 1 and len(v0) != nfase:
        v0 *= nfase
    v = v0.copy()
    if len(Xass0) != 1 and len(Xass0) != nfase:
        Xass0 *= nfase
    Xass = Xass0.copy()

    for i, state in enumerate(equilibrium):
        lnphi[i], v[i], Xass[i] = model.logfugef_aux(X[i], temp_aux, P, state,
                                                     v[i], Xass[i])
        if np.any(np.isnan(lnphi[i])) or np.isnan(v[i]):
            lnphi[i], v[i], Xass[i] = model.logfugef_aux(X[i], temp_aux, P, state)

    lnK = lnphi[0] - lnphi[1:]
    K = np.exp(lnK)

    x = betatetha
    method = "ASS"
    while error > K_tol and itacc < nacc:
        ittotal += 1
        it += 1
        lnK_old = lnK.copy()

        ef = 1.
        ex = 1.
        itin = 0
        while ef > 1e-8 and ex > 1e-8 and itin < 30:
            itin += 1
            f, jac, Kexp, Xref = multiflash_obj(x, z, K)
            dx = np.linalg.solve(jac, -f)
            x += dx
            ef = np.linalg.norm(f)
            ex = np.linalg.norm(dx)
            x[x > tetha_max] = tetha_max

        x[x <= 1e-10] = 0.
        beta, tetha = np.array_split(x, 2)
        beta /= beta.sum()

        # Update compositions
        X[0] = Xref
        X[1:] = Xref*Kexp.T
        X = np.abs(X)
        X = (X.T/X.sum(axis=1)).T

        # Update fugacity coefficient
        for i, state in enumerate(equilibrium):
            lnphi[i], v[i], Xass[i] = model.logfugef_aux(X[i], temp_aux, P,
                                                         state, v[i], Xass[i])
            if np.any(np.isnan(lnphi[i])) or np.isnan(v[i]):
                lnphi[i], v[i], Xass[i] = model.logfugef_aux(X[i], temp_aux, P, state)

        lnK = lnphi[0] - lnphi[1:]
        error = np.sum((lnK - lnK_old)**2)

        # Accelerated succesive sustitution
        if it == (n-3):
            lnK3 = lnK.flatten()
        elif it == (n-2):
            lnK2 = lnK.flatten()
        elif it == (n-1):
            lnK1 = lnK.flatten()
        elif it == n:
            it = 0
            itacc += 1
            lnKf = lnK.flatten()
            dacc = gdem(lnKf, lnK1, lnK2, lnK3).reshape(lnK.shape)
            lnK += dacc

        K = np.exp(lnK)

    if error > K_tol and itacc == nacc and ef < 1e-8 and np.all(tetha == 0):
        fobj = gibbs_obj
        jac = True
        hess = None
        method = 'BFGS'
        global vg, Xassg
        vg = v.copy()
        Xassg = Xass.copy()
        z_notzero = np.nonzero(z)[0]
        n_nonzero = len(z_notzero)

        ind0 = (X.copy()[:, z_notzero].T*beta).T[1:].flatten()

        ind1 = minimize(fobj, ind0, args=(equilibrium, z, z_notzero, n_nonzero,
                        temp_aux, P, model), jac=jac, method=method,
                        hess=hess)

        if ind1.success:
            method = "Gibbs_minimization"
            v = vg.copy()
            Xass = Xassg.copy()

            ittotal += ind1.nit
            error = np.linalg.norm(ind1.jac)
            nc = model.nc

            ind = 1.*ind1.x.reshape(nfase-1, n_nonzero)
            dep = z[z_notzero] - ind.sum(axis=0)

            X = np.zeros((nfase, nc))
            X[1:, z_notzero] = ind
            X[0, z_notzero] = dep
            beta = X.sum(axis=1)
            X = (X.T/beta).T
            # updating volume roots for founded equilibria compositions
            for i, state in enumerate(equilibrium):
                rho, Xass[i] = model.density_aux(X[i], temp_aux, P, state,
                                                 rho0=1/v[i], Xass0=Xass[i])
                v[i] = 1./rho

    if full_output:
        sol = {'T': T, 'P': P, 'error_outer': error, 'error_inner': ef,
               'iter': ittotal, 'beta': beta, 'tetha': tetha, 'X': X, 'v': v,
               'Xass': Xass, 'states': equilibrium, 'method': method}
        out = EquilibriumResult(sol)
        return out

    return X, beta, tetha



from __future__ import division, print_function, absolute_import
import numpy as np
from .a1s_monomer import a1s, da1s_dxhi00, d2a1s_dxhi00, d3a1s_dxhi00
from .B_monomer import B, dB_dxhi00, d2B_dxhi00, d3B_dxhi00
from .a1s_monomer import da1s_dx_dxhi00_dxxhi, da1s_dx_d2xhi00_dxxhi
from .B_monomer import dB_dx_dxhi00_dxxhi, dB_dx_d2xhi00_dxxhi


def a1sB(xhi00, xhix, xhix_vec, xm, Ilam, Jlam, cictes, a1vdw, a1vdw_cte):
    a1 = a1s(xhi00, xhix_vec, xm, cictes, a1vdw)
    b = B(xhi00, xhix, xm, Ilam, Jlam, a1vdw_cte)
    return a1 + b


def da1sB_dxhi00(xhi00, xhix, xhix_vec, xm, Ilam, Jlam, cictes, a1vdw,
                 a1vdw_cte, dxhix_dxhi00):

    a1, da1 = da1s_dxhi00(xhi00, xhix_vec, xm, cictes, a1vdw, dxhix_dxhi00)
    b, db = dB_dxhi00(xhi00, xhix, xm, Ilam, Jlam, a1vdw_cte, dxhix_dxhi00)

    return a1 + b, da1 + db


def d2a1sB_dxhi00(xhi00, xhix, xhix_vec, xm, Ilam, Jlam, cictes,
                  a1vdw, a1vdw_cte, dxhix_dxhi00):
    a1, da1, d2a1 = d2a1s_dxhi00(xhi00, xhix_vec, xm, cictes, a1vdw,
                                 dxhix_dxhi00)
    b, db, d2b = d2B_dxhi00(xhi00, xhix, xm, Ilam, Jlam, a1vdw_cte,
                            dxhix_dxhi00)
    return a1 + b, da1 + db, d2a1 + d2b


def d3a1sB_dxhi00(xhi00, xhix, xhix_vec, xm, Ilam, Jlam, cictes, a1vdw,
                  a1vdw_cte, dxhix_dxhi00):
    a1, da1, d2a1, d3a1 = d3a1s_dxhi00(xhi00, xhix_vec, xm, cictes, a1vdw,
                                       dxhix_dxhi00)
    b, db, d2b, d3b = d3B_dxhi00(xhi00, xhix, xm, Ilam, Jlam, a1vdw_cte,
                                 dxhix_dxhi00)
    return a1 + b, da1 + db, d2a1 + d2b, d3a1 + d3b


def da1sB_dx_dxhi00_dxxhi(xhi00, xhix, xhix_vec, xm, ms, I_ij,
                          J_ij, cctesij, a1vdwij, a1vdw_cteij,
                          dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00):
    out = da1s_dx_dxhi00_dxxhi(xhi00, xhix_vec, xm, ms, cctesij, a1vdwij,
                               dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00)
    a1, da1, da1x, da1xxhi = out
    out = dB_dx_dxhi00_dxxhi(xhi00, xhix, xm, ms, I_ij, J_ij, a1vdw_cteij,
                             dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00)
    b, db, dbx, dbxxhi = out
    return a1+b, da1+db, da1x+dbx, da1xxhi+dbxxhi


def da1sB_dx_d2xhi00_dxxhi(xhi00, xhix, xhix_vec, xm, ms, I_ij,
                           J_ij, cctesij, a1vdwij, a1vdw_cteij,
                           dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00):
    out = da1s_dx_d2xhi00_dxxhi(xhi00, xhix_vec, xm, ms, cctesij, a1vdwij,
                                dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00)
    a1, da1, d2a1, da1x, da1xxhi = out
    out = dB_dx_d2xhi00_dxxhi(xhi00, xhix, xm, ms, I_ij, J_ij, a1vdw_cteij,
                              dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00)
    b, db, d2b, dbx, dbxxhi = out
    return a1+b, da1+db, d2a1+d2b, da1x+dbx, da1xxhi+dbxxhi


def a1sB_eval(xhi00, xhix, xhix_vec, xm, I_lambdasij, J_lambdasij, cctesij,
              a1vdwij, a1vdw_cteij):

    # laij, lrij, larij = lambdas
    cctes_laij, cctes_lrij, cctes_2laij, cctes_2lrij, cctes_larij = cctesij
    a1vdw_laij, a1vdw_lrij, a1vdw_2laij, a1vdw_2lrij, a1vdw_larij = a1vdwij
    I_la, I_lr, I_2la, I_2lr, I_lar = I_lambdasij
    J_la, J_lr, J_2la, J_2lr, J_lar = J_lambdasij

    a1sb_a = a1sB(xhi00, xhix, xhix_vec, xm, I_la, J_la, cctes_laij,
                  a1vdw_laij, a1vdw_cteij)
    a1sb_r = a1sB(xhi00, xhix, xhix_vec, xm, I_lr, J_lr, cctes_lrij,
                  a1vdw_lrij, a1vdw_cteij)
    a1sb_2a = a1sB(xhi00, xhix, xhix_vec, xm, I_2la, J_2la, cctes_2laij,
                   a1vdw_2laij, a1vdw_cteij)
    a1sb_2r = a1sB(xhi00, xhix, xhix_vec, xm, I_2lr, J_2lr, cctes_2lrij,
                   a1vdw_2lrij, a1vdw_cteij)
    a1sb_ar = a1sB(xhi00, xhix, xhix_vec, xm, I_lar, J_lar, cctes_larij,
                   a1vdw_larij, a1vdw_cteij)

    a1sb_a1 = np.array([a1sb_a, a1sb_r])
    a1sb_a2 = np.array([a1sb_2a, a1sb_ar, a1sb_2r])
    return a1sb_a1, a1sb_a2


def da1sB_dxhi00_eval(xhi00, xhix, xhix_vec, xm, I_lambdasij, J_lambdasij,
                      cctesij, a1vdwij, a1vdw_cteij, dxhix_dxhi00):

    # laij, lrij, larij = lambdas
    cctes_laij, cctes_lrij, cctes_2laij, cctes_2lrij, cctes_larij = cctesij
    a1vdw_laij, a1vdw_lrij, a1vdw_2laij, a1vdw_2lrij, a1vdw_larij = a1vdwij
    I_la, I_lr, I_2la, I_2lr, I_lar = I_lambdasij
    J_la, J_lr, J_2la, J_2lr, J_lar = J_lambdasij

    a1sb_a, da1sb_a = da1sB_dxhi00(xhi00, xhix, xhix_vec, xm, I_la, J_la,
                                   cctes_laij, a1vdw_laij, a1vdw_cteij,
                                   dxhix_dxhi00)

    a1sb_r, da1sb_r = da1sB_dxhi00(xhi00, xhix, xhix_vec, xm, I_lr, J_lr,
                                   cctes_lrij, a1vdw_lrij, a1vdw_cteij,
                                   dxhix_dxhi00)

    a1sb_2a, da1sb_2a = da1sB_dxhi00(xhi00, xhix, xhix_vec, xm, I_2la,
                                     J_2la, cctes_2laij, a1vdw_2laij,
                                     a1vdw_cteij, dxhix_dxhi00)

    a1sb_2r, da1sb_2r = da1sB_dxhi00(xhi00, xhix, xhix_vec, xm, I_2lr,
                                     J_2lr, cctes_2lrij, a1vdw_2lrij,
                                     a1vdw_cteij, dxhix_dxhi00)

    a1sb_ar, da1sb_ar = da1sB_dxhi00(xhi00, xhix, xhix_vec, xm, I_lar,
                                     J_lar, cctes_larij, a1vdw_larij,
                                     a1vdw_cteij, dxhix_dxhi00)

    a1sb_a1 = np.array([[a1sb_a, a1sb_r],
                        [da1sb_a, da1sb_r]])
    a1sb_a2 = np.array([[a1sb_2a, a1sb_ar, a1sb_2r],
                       [da1sb_2a, da1sb_ar, da1sb_2r]])

    return a1sb_a1, a1sb_a2


def d2a1sB_dxhi00_eval(xhi00, xhix, xhix_vec, xm, I_lambdasij, J_lambdasij,
                       cctesij, a1vdwij, a1vdw_cteij, dxhix_dxhi00):

    cctes_laij, cctes_lrij, cctes_2laij, cctes_2lrij, cctes_larij = cctesij
    a1vdw_laij, a1vdw_lrij, a1vdw_2laij, a1vdw_2lrij, a1vdw_larij = a1vdwij
    I_la, I_lr, I_2la, I_2lr, I_lar = I_lambdasij
    J_la, J_lr, J_2la, J_2lr, J_lar = J_lambdasij

    out = d2a1sB_dxhi00(xhi00, xhix, xhix_vec, xm,  I_la, J_la,
                        cctes_laij, a1vdw_laij, a1vdw_cteij, dxhix_dxhi00)
    a1sb_a, da1sb_a, d2a1sb_a = out

    out = d2a1sB_dxhi00(xhi00, xhix, xhix_vec, xm, I_lr, J_lr,
                        cctes_lrij, a1vdw_lrij, a1vdw_cteij, dxhix_dxhi00)
    a1sb_r, da1sb_r, d2a1sb_r = out

    out = d2a1sB_dxhi00(xhi00, xhix, xhix_vec, xm,  I_2la, J_2la,
                        cctes_2laij, a1vdw_2laij, a1vdw_cteij, dxhix_dxhi00)
    a1sb_2a, da1sb_2a, d2a1sb_2a = out

    out = d2a1sB_dxhi00(xhi00, xhix, xhix_vec, xm,  I_2lr, J_2lr,
                        cctes_2lrij, a1vdw_2lrij, a1vdw_cteij, dxhix_dxhi00)
    a1sb_2r, da1sb_2r, d2a1sb_2r = out

    out = d2a1sB_dxhi00(xhi00, xhix, xhix_vec, xm,  I_lar, J_lar,
                        cctes_larij, a1vdw_larij, a1vdw_cteij, dxhix_dxhi00)
    a1sb_ar, da1sb_ar, d2a1sb_ar = out

    a1sb_a1 = np.array([[a1sb_a, a1sb_r],
                        [da1sb_a, da1sb_r],
                       [d2a1sb_a, d2a1sb_r]])

    a1sb_a2 = np.array([[a1sb_2a, a1sb_ar, a1sb_2r],
                       [da1sb_2a, da1sb_ar, da1sb_2r],
                       [d2a1sb_2a, d2a1sb_ar, d2a1sb_2r]])
    return a1sb_a1, a1sb_a2


def d3a1sB_dxhi00_eval(xhi00, xhix, xhix_vec, xm, I_lambdasij, J_lambdasij,
                       cctesij, a1vdwij, a1vdw_cteij, dxhix_dxhi00):

    # laij, lrij, larij = lambdas
    cctes_laij, cctes_lrij, cctes_2laij, cctes_2lrij, cctes_larij = cctesij
    a1vdw_laij, a1vdw_lrij, a1vdw_2laij, a1vdw_2lrij, a1vdw_larij = a1vdwij
    I_la, I_lr, I_2la, I_2lr, I_lar = I_lambdasij
    J_la, J_lr, J_2la, J_2lr, J_lar = J_lambdasij

    out = d3a1sB_dxhi00(xhi00, xhix, xhix_vec, xm, I_la, J_la, cctes_laij,
                        a1vdw_laij, a1vdw_cteij, dxhix_dxhi00)
    a1sb_a, da1sb_a, d2a1sb_a, d3a1sb_a = out

    out = d3a1sB_dxhi00(xhi00, xhix, xhix_vec, xm, I_lr, J_lr, cctes_lrij,
                        a1vdw_lrij, a1vdw_cteij, dxhix_dxhi00)
    a1sb_r, da1sb_r, d2a1sb_r, d3a1sb_r = out

    out = d3a1sB_dxhi00(xhi00, xhix, xhix_vec, xm, I_2la, J_2la,
                        cctes_2laij, a1vdw_2laij, a1vdw_cteij, dxhix_dxhi00)
    a1sb_2a, da1sb_2a, d2a1sb_2a, d3a1sb_2a = out

    out = d3a1sB_dxhi00(xhi00, xhix, xhix_vec, xm, I_2lr, J_2lr,
                        cctes_2lrij, a1vdw_2lrij, a1vdw_cteij, dxhix_dxhi00)
    a1sb_2r, da1sb_2r, d2a1sb_2r, d3a1sb_2r = out

    out = d3a1sB_dxhi00(xhi00, xhix, xhix_vec, xm, I_lar, J_lar,
                        cctes_larij, a1vdw_larij, a1vdw_cteij, dxhix_dxhi00)
    a1sb_ar, da1sb_ar, d2a1sb_ar, d3a1sb_ar = out

    a1sb_a1 = np.array([[a1sb_a, a1sb_r],
                        [da1sb_a, da1sb_r],
                       [d2a1sb_a, d2a1sb_r],
                       [d3a1sb_a, d3a1sb_r]])

    a1sb_a2 = np.array([[a1sb_2a, a1sb_ar, a1sb_2r],
                       [da1sb_2a, da1sb_ar, da1sb_2r],
                       [d2a1sb_2a, d2a1sb_ar, d2a1sb_2r],
                       [d3a1sb_2a, d3a1sb_ar, d3a1sb_2r]])

    return a1sb_a1, a1sb_a2


def da1sB_dx_dxhi00_dxxhi_eval(xhi00, xhix, xhix_vec, xm, ms, I_lambdasij,
                               J_lambdasij, cctesij, a1vdwij, a1vdw_cteij,
                               dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00):

    cctes_laij, cctes_lrij, cctes_2laij, cctes_2lrij, cctes_larij = cctesij
    a1vdw_laij, a1vdw_lrij, a1vdw_2laij, a1vdw_2lrij, a1vdw_larij = a1vdwij
    I_laij, I_lrij, I_2laij, I_2lrij, I_larij = I_lambdasij
    J_laij, J_lrij, J_2laij, J_2lrij, J_larij = J_lambdasij

    out_la = da1sB_dx_dxhi00_dxxhi(xhi00, xhix, xhix_vec, xm, ms, I_laij,
                                   J_laij, cctes_laij, a1vdw_laij, a1vdw_cteij,
                                   dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00)
    a1sb_a, da1sb_a, da1sb_ax, da1sb_axxhi = out_la
    out_lr = da1sB_dx_dxhi00_dxxhi(xhi00, xhix, xhix_vec, xm, ms, I_lrij,
                                   J_lrij, cctes_lrij, a1vdw_lrij, a1vdw_cteij,
                                   dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00)
    a1sb_r, da1sb_r, da1sb_rx, da1sb_rxxhi = out_lr
    out_2la = da1sB_dx_dxhi00_dxxhi(xhi00, xhix, xhix_vec, xm, ms, I_2laij,
                                    J_2laij, cctes_2laij, a1vdw_2laij,
                                    a1vdw_cteij, dxhix_dxhi00, dxhix_dx,
                                    dxhix_dx_dxhi00)
    a1sb_2a, da1sb_2a, da1sb_2ax, da1sb_2axxhi = out_2la
    out_2lr = da1sB_dx_dxhi00_dxxhi(xhi00, xhix, xhix_vec, xm, ms, I_2lrij,
                                    J_2lrij, cctes_2lrij, a1vdw_2lrij,
                                    a1vdw_cteij, dxhix_dxhi00, dxhix_dx,
                                    dxhix_dx_dxhi00)
    a1sb_2r, da1sb_2r, da1sb_2rx, da1sb_2rxxhi = out_2lr
    out_lar = da1sB_dx_dxhi00_dxxhi(xhi00, xhix, xhix_vec, xm, ms, I_larij,
                                    J_larij, cctes_larij, a1vdw_larij,
                                    a1vdw_cteij, dxhix_dxhi00, dxhix_dx,
                                    dxhix_dx_dxhi00)
    a1sb_ar, da1sb_ar, da1sb_arx, da1sb_arxxhi = out_lar

    a1sb_a1 = np.array([[a1sb_a, a1sb_r],
                        [da1sb_a, da1sb_r]])

    a1sb_a2 = np.array([[a1sb_2a, a1sb_ar, a1sb_2r],
                       [da1sb_2a, da1sb_ar, da1sb_2r]])

    a1sb_a1x = np.array([da1sb_ax, da1sb_rx])
    a1sb_a2x = np.array([da1sb_2ax, da1sb_arx, da1sb_2rx])

    a1sb_a1xxhi = np.array([da1sb_axxhi, da1sb_rxxhi])
    a1sb_a2xxhi = np.array([da1sb_2axxhi, da1sb_arxxhi, da1sb_2rxxhi])

    return a1sb_a1, a1sb_a2, a1sb_a1x, a1sb_a2x, a1sb_a1xxhi, a1sb_a2xxhi


def da1sB_dx_d2xhi00_dxxhi_eval(xhi00, xhix, xhix_vec, xm, ms, I_lambdasij,
                                J_lambdasij, cctesij, a1vdwij, a1vdw_cteij,
                                dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00):

    cctes_laij, cctes_lrij, cctes_2laij, cctes_2lrij, cctes_larij = cctesij
    a1vdw_laij, a1vdw_lrij, a1vdw_2laij, a1vdw_2lrij, a1vdw_larij = a1vdwij
    I_laij, I_lrij, I_2laij, I_2lrij, I_larij = I_lambdasij
    J_laij, J_lrij, J_2laij, J_2lrij, J_larij = J_lambdasij

    out_la = da1sB_dx_d2xhi00_dxxhi(xhi00, xhix, xhix_vec, xm, ms, I_laij,
                                    J_laij, cctes_laij, a1vdw_laij,
                                    a1vdw_cteij, dxhix_dxhi00, dxhix_dx,
                                    dxhix_dx_dxhi00)
    a1sb_a, da1sb_a, d2a1sb_a, da1sb_ax, da1sb_axxhi = out_la
    out_lr = da1sB_dx_d2xhi00_dxxhi(xhi00, xhix, xhix_vec, xm, ms, I_lrij,
                                    J_lrij, cctes_lrij, a1vdw_lrij,
                                    a1vdw_cteij, dxhix_dxhi00, dxhix_dx,
                                    dxhix_dx_dxhi00)
    a1sb_r, da1sb_r, d2a1sb_r, da1sb_rx, da1sb_rxxhi = out_lr
    out_2la = da1sB_dx_d2xhi00_dxxhi(xhi00, xhix, xhix_vec, xm, ms, I_2laij,
                                     J_2laij, cctes_2laij, a1vdw_2laij,
                                     a1vdw_cteij, dxhix_dxhi00, dxhix_dx,
                                     dxhix_dx_dxhi00)
    a1sb_2a, da1sb_2a, d2a1sb_2a, da1sb_2ax, da1sb_2axxhi = out_2la
    out_2lr = da1sB_dx_d2xhi00_dxxhi(xhi00, xhix, xhix_vec, xm, ms, I_2lrij,
                                     J_2lrij, cctes_2lrij, a1vdw_2lrij,
                                     a1vdw_cteij, dxhix_dxhi00, dxhix_dx,
                                     dxhix_dx_dxhi00)
    a1sb_2r, da1sb_2r, d2a1sb_2r, da1sb_2rx, da1sb_2rxxhi = out_2lr
    out_lar = da1sB_dx_d2xhi00_dxxhi(xhi00, xhix, xhix_vec, xm, ms, I_larij,
                                     J_larij, cctes_larij, a1vdw_larij,
                                     a1vdw_cteij, dxhix_dxhi00, dxhix_dx,
                                     dxhix_dx_dxhi00)
    a1sb_ar, da1sb_ar, d2a1sb_ar, da1sb_arx, da1sb_arxxhi = out_lar

    a1sb_a1 = np.array([[a1sb_a, a1sb_r],
                        [da1sb_a, da1sb_r],
                        [d2a1sb_a, d2a1sb_r]])

    a1sb_a2 = np.array([[a1sb_2a, a1sb_ar, a1sb_2r],
                       [da1sb_2a, da1sb_ar, da1sb_2r],
                       [d2a1sb_2a, d2a1sb_ar, d2a1sb_2r]])

    a1sb_a1x = np.array([da1sb_ax, da1sb_rx])
    a1sb_a2x = np.array([da1sb_2ax, da1sb_arx, da1sb_2rx])

    a1sb_a1xxhi = np.array([da1sb_axxhi, da1sb_rxxhi])
    a1sb_a2xxhi = np.array([da1sb_2axxhi, da1sb_arxxhi, da1sb_2rxxhi])

    return a1sb_a1, a1sb_a2, a1sb_a1x, a1sb_a2x, a1sb_a1xxhi, a1sb_a2xxhi


def x0lambda_eval(x0, la, lr, lar, laij, lrij, larij, diag_index):
    x0la = x0**laij
    x0lr = x0**lrij
    x02la = x0**(2*laij)
    x02lr = x0**(2*lrij)
    x0lar = x0**larij

    # To be used for a1 and a2 of monomer
    x0_a1 = np.array([x0la, -x0lr])
    x0_a2 = np.array([x02la, -2*x0lar, x02lr])

    # To be used in g1 and g2 of chain
    x0_g1 = np.array([la * x0la[diag_index], -lr*x0lr[diag_index]])
    x0_g2 = np.array([la * x02la[diag_index], -lar*x0lar[diag_index],
                     lr * x02lr[diag_index]])

    return x0_a1, x0_a2, x0_g1, x0_g2



from __future__ import division, print_function, absolute_import
import numpy as np


# Hard sphere Eq A6
def ahs(xhi):
    xhi0, xhi1, xhi2, xhi3 = xhi
    xhi3_1 = 1-xhi3
    xhi23 = xhi2**3

    a = (xhi23/xhi3**2 - xhi0) * np.log(xhi3_1)
    a += 3 * xhi1 * xhi2 / xhi3_1
    a += xhi23 / xhi3 / xhi3_1**2
    a /= xhi0
    return a


def dahs_dxhi00(xhi, dxhi_dxhi00):
    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0, dxhi1, dxhi2, dxhi3 = dxhi_dxhi00
    xhi3_1 = 1-xhi3
    xhi3_12 = xhi3_1**2
    xhi3_13 = xhi3_12*xhi3_1
    xhi23 = xhi2**3

    a = (xhi23/xhi3**2 - xhi0) * np.log(xhi3_1)
    a += 3 * xhi1 * xhi2 / xhi3_1
    a += xhi23 / xhi3 / xhi3_12
    a /= xhi0

    da = xhi2 * dxhi2**2 * (-3 + xhi3)
    da += 3 * dxhi1 * dxhi2 * -xhi3_1
    da -= dxhi0 * dxhi3 * xhi3_12
    da /= -xhi3_13 * dxhi0
    return np.array([a, da])


def d2ahs_dxhi00(xhi, dxhi_dxhi00):

    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0, dxhi1, dxhi2, dxhi3 = dxhi_dxhi00
    xhi3_1 = 1-xhi3
    xhi3_12 = xhi3_1**2
    xhi3_13 = xhi3_12*xhi3_1
    xhi3_14 = xhi3_13*xhi3_1
    xhi23 = xhi2**3
    dxhi22 = dxhi2**2
    dxhi23 = dxhi22*dxhi2

    a = (xhi23/xhi3**2 - xhi0) * np.log(xhi3_1)
    a += 3 * xhi1 * xhi2 / xhi3_1
    a += xhi23 / xhi3 / xhi3_12
    a /= xhi0

    da = xhi2 * dxhi22 * (-3 + xhi3)
    da += 3 * dxhi1 * dxhi2 * -xhi3_1
    da -= dxhi0 * dxhi3 * xhi3_12
    da /= -xhi3_13 * dxhi0

    d2a = -6 * dxhi1 * dxhi2 * dxhi3 * -xhi3_1
    d2a += dxhi0 * dxhi3**2 * xhi3_12
    d2a += dxhi23 * (3 + xhi3 * (4 - xhi3))
    d2a /= xhi3_14 * dxhi0
    return np.array([a, da, d2a])


def dahs_dx(xhi, dxhi_dx):

    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0x, dxhi1x, dxhi2x, dxhi3x = dxhi_dx

    xhi3_1 = 1-xhi3
    xhi3_12 = xhi3_1**2
    xhi3_13 = xhi3_12*xhi3_1
    log3 = np.log(xhi3_1)

    xhi02 = xhi0**2
    xhi22 = xhi2**2
    xhi23 = xhi22*xhi2

    xhi32 = xhi3**2
    xhi33 = xhi32*xhi3

    a = (xhi23/xhi32 - xhi0) * log3
    a += 3 * xhi1 * xhi2 / xhi3_1
    a += xhi23 / xhi3 / xhi3_12
    a /= xhi0

    dax1 = -3*xhi0*xhi22 * (log3 + xhi3/xhi3_12)
    dax1 *= dxhi2x / xhi32

    dax2 = (xhi2 * dxhi0x - xhi0 * dxhi2x) * -xhi3_1
    dax2 += xhi0 * xhi2 * dxhi3x
    dax2 *= 3. * xhi1/xhi3_1

    dax2 += 3 * xhi0 * xhi2 * dxhi1x
    dax2 += xhi02 * dxhi3x
    dax2 /= -xhi3_1

    dax3 = log3 * dxhi0x - xhi0 * (-2 + xhi3)*dxhi3x/xhi3_12
    dax3 *= xhi3
    dax3 += 2*log3*xhi0*dxhi3x
    dax3 += xhi32 * (-xhi3_1*dxhi0x + 2*xhi0*dxhi3x)/-xhi3_13
    dax3 *= (xhi23/xhi33)

    dax = dax1 + dax2 + dax3
    dax /= -xhi02

    return a, dax


def dahs_dxxhi(xhi, dxhi_dxhi00, dxhi_dx):

    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0, dxhi1, dxhi2, dxhi3 = dxhi_dxhi00
    dxhi0x, dxhi1x, dxhi2x, dxhi3x = dxhi_dx

    xhi3_1 = 1-xhi3
    xhi3_12 = xhi3_1**2
    xhi3_13 = xhi3_12*xhi3_1
    log3 = np.log(xhi3_1)

    xhi02 = xhi0**2
    xhi22 = xhi2**2
    xhi23 = xhi22*xhi2

    xhi32 = xhi3**2
    xhi33 = xhi32*xhi3

    a = (xhi23/xhi32 - xhi0) * log3
    a += 3 * xhi1 * xhi2 / xhi3_1
    a += xhi23 / xhi3 / xhi3_12
    a /= xhi0

    da = xhi2 * dxhi2**2 * (-3 + xhi3)
    da += 3 * dxhi1 * dxhi2 * -xhi3_1
    da -= dxhi0 * dxhi3 * xhi3_12
    da /= -xhi3_13 * dxhi0

    dax1 = -3*xhi0*xhi22 * (log3 + xhi3/xhi3_12)
    dax1 *= dxhi2x / xhi32

    dax2 = (xhi2 * dxhi0x - xhi0 * dxhi2x) * -xhi3_1
    dax2 += xhi0 * xhi2 * dxhi3x
    dax2 *= 3. * xhi1/xhi3_1

    dax2 += 3 * xhi0 * xhi2 * dxhi1x
    dax2 += xhi02 * dxhi3x
    dax2 /= -xhi3_1

    dax3 = log3 * dxhi0x - xhi0 * (-2 + xhi3)*dxhi3x/xhi3_12
    dax3 *= xhi3
    dax3 += 2*log3*xhi0*dxhi3x
    dax3 += xhi32 * (-xhi3_1*dxhi0x + 2*xhi0*dxhi3x)/-xhi3_13
    dax3 *= (xhi23/xhi33)

    dax = dax1 + dax2 + dax3
    dax /= -xhi02

    return np.array([a, da]), dax



from __future__ import division, print_function, absolute_import
from scipy.optimize import brentq, newton
from .EquilibriumResult import EquilibriumResult
from warnings import warn


def fobj_tsat(T, P, saft):
    global vl, vv
    global Xassl, Xassv
    temp_aux = saft.temperature_aux(T)
    lnphiv, vv, Xassv = saft.logfug_aux(temp_aux, P, 'V', vv, Xassv)
    lnphil, vl, Xassl = saft.logfug_aux(temp_aux, P, 'L', vl, Xassl)
    FO = lnphiv - lnphil
    return FO


def tsat(saft, P, T0=None, Tbounds=None, v0=[None, None], Xass0=[None, None],
         full_output=False):

    if saft.critical:
        if P >= saft.Pc:
            warn('Pressure is greater than critical pressure, returning critical point')
            if full_output:
                dict = {'T': saft.Tc, 'P': saft.Pc, 'vl': 1./saft.rhoc,
                        'vv': 1./saft.rhoc, 'Xassl': Xass0[0],
                        'Xassv': Xass0[1], 'success': False, 'iterations': 0}
                out = EquilibriumResult(dict)
            else:
                out = saft.Tc, 1./saft.rhoc, 1./saft.rhoc
            return out

    bool1 = T0 is None
    bool2 = Tbounds is None

    if bool1 and bool2:
        raise Exception('You must provide either Tbounds or T0')

    global Xassl, Xassv
    Xassl, Xassv = Xass0

    global vl, vv
    vl, vv = v0
    if not bool1:
        sol = newton(fobj_tsat, x0=T0, args=(P, saft),
                     full_output=True)
    elif not bool2:
        sol = brentq(fobj_tsat, Tbounds[0], Tbounds[1], args=(P, saft),
                     full_output=True)

    Tsat = sol[0]
    if full_output:
        dict = {'T': Tsat, 'P': P, 'vl': vl, 'vv': vv, 'Xassl': Xassl,
                'Xassv': Xassv, 'success': sol[1].converged,
                'iterations': sol[1].iterations}
        out = EquilibriumResult(dict)
    else:
        out = (Tsat, vl, vv)

    return out
